""" Chebyshev propagator for classical particle dynamics """

import numpy as np
import sympy as sp
from scipy.special import jn, iv
from propagators import PolynomialPropagator
from liouville import LiouvillePowersQP

def cheb_coef(bf, nterms, alpha):
    coef = bf(range(nterms), np.full(nterms, alpha))
    coef[1:] = coef[1:]*2.
    return coef

def besf(fpprec, signed, prec=None):
    if fpprec == 'fixed':
        bessel = iv if signed else jn
    elif fpprec == 'multi':
        fargs = [] if prec is None else [prec]
        bess = sp.besseli if signed else sp.besselj
        def bessel(orders, args):
            bfun = [bess(o, a).evalf(*fargs) for o, a in zip(orders, args)]
            return np.array(bfun)
    else:
        bessel = None
    return bessel


class Chebyshev(PolynomialPropagator):
    """ Chebyshev propagator for classical particle dynamics """
    name = 'chebyshev'
    def __init__(self, delta=1.0, lambda_min=None, nterms=None, signed=False,
                 liouville=LiouvillePowersQP, liou_params={}, fpprec='fixed',
                 prec=None, **kwargs):
        super().__init__(**kwargs)

        self.delta = delta
        alpha = self.delta*self.tstep/2.
        self.nterms = int(np.ceil(alpha)) if nterms is None else nterms
        self.efficiency = alpha/self.nterms
        self.lambda_min = -self.delta/2. if lambda_min is None else lambda_min
        self.scale = 2./self.delta
        self.shift = -(2.*self.lambda_min/self.delta+1.)
        bf = besf(fpprec, signed, prec=prec)
        sign = -1 if signed else 1
        exp = np.exp if fpprec == 'fixed' else lambda x: sp.exp(x).evalf(prec)

        if self.substep is not None:
            self.cheb_coef = []
            for stime in self.subtime:
                alpha = self.delta*stime/2.
                scaling = exp((self.lambda_min+self.delta/2.0)*stime)
                coef = cheb_coef(bf, self.nterms, alpha)*scaling
                self.cheb_coef.append(coef)
        else:
            scaling = exp((self.lambda_min+self.delta/2.0)*self.tstep)
            self.cheb_coef = cheb_coef(bf, self.nterms, alpha)*scaling

        self.iln = liouville(self.nterms, self.syst, scale=self.scale,
                             shift=self.shift, recursion='chebyshev',
                             sign=sign, fpprec=fpprec, **liou_params)
        self.coeffs = self.cheb_coef
